!+
! Module: epnoise_mod
!
! Authors:  Aitor Ibarra - XMM-Newton SOC
!
! Description: Algorithm to reject soft X-ray noise in the EPIC-pn camera
!
! Comments: 
! This task has two different parts:
! 1.- Identify noisy frames and write keywords containing suggestions for subsequent
!       filtering. The input parameter for this part is the output of epframes and BADPIX
!       extension, written by badpixfind.
! 2.- Filter the final event list using the threshold calculated in the previous step
!       and update the ONTIME and LIVETIME keywords. The input parameter for this step is the
!       output of epframes, updated with all the columns (NEVT_FRM) and keywords (NEVT_CUT), and
!       the final event list.
!
! Step 1:
! epnoise appends or updates the column NEVT_FRM containing the number of events per frame and CCD with
! PHA values in [20..iamx], and writes keywords containing suggestions for subsequent filtering.
! Pixels which are affected by bright celestial sources in the [20..iamx] range or which are noisy
! are removed through a mask generation.
! To create the mask for removing bright sources, epnoise calculate the median of the full image and apply a
! cut using the sigmacut parameter.
! Then a mask for badpixel is created and added to the previous mask.
! The parameter "savemasks" writes to disk the masks of all CCDs.
!
! Step 2:
! Filter the final event list using the NEVT_CUT threshold. Update the STDGTI extension adding the gaps
! corresponding to the events that have been removed. Then, update the ONTIME and LIVETIME keywords.
! 
!
! $Id: epnoise_mod.f90,v 1.2 2011/09/13 14:07:42 aibarra Exp $
!-
module epnoise_mod

  use dal
  use param  
  use errorHandling
  use epnoisemask_mod
  use removenoisyframes_mod
  use eventAttributes
  use eventAttributes ! in saslib/sasdal. For checking which flags have

  implicit none 

  private 

  integer, parameter :: maxfrm   = 200000000  ,& ! 1150 cts/s telem. limit
                                                 ! and 48 hrs max exp. and
                                                 ! min. one event per frame
                        maxbuf   = 12800      ,& ! buffer size = 64*200
                        maxcol   = 10         ,&
                        maxhst   = 1000       ,&
                        nxdim    = 64         ,&
                        nydim    = 200        ,&
                        mxdim    = 6*nxdim+6  ,&
                        mydim    = 2*nydim+2
  logical, parameter  :: ldebug  = .true. ! .FALSE.        ! toggle NOISY verbosity for debug purposes at compile time

  integer(kind=Int32) :: nfrm,nrows,iamx,i4map(1:nxdim,1:nydim)
  integer(kind=Int32), dimension(:), allocatable :: ind
  integer(kind=Int16), dimension(:), pointer :: energyCol, rawXCol,rawYCol
  integer(kind=Int32), dimension(:), pointer :: frameCol

  integer(kind=Int16) :: kx, ky, noise(maxbuf)
  integer(kind=Int32) :: total_array(1:mxdim,1:mydim) = -1
  integer(kind=Int32) :: i4mask(1:nxdim,1:nydim)
  integer(kind=Int32) :: i4hst(maxhst), i4ac1(maxhst), i4ac2(maxhst),&
                         i4ac3(maxhst), i4ac4(maxhst)

  public epnoise

contains 


  !---------------------------------------------------------------------
  subroutine epnoise
  !---------------------------------------------------------------------
    implicit none

    integer :: nfiles,i
    type(DataSetT)      :: outmasks
    type(ArrayT)        :: total
    character(len=8)    :: outMapExtension
    character(len=1024) :: outmapset = "epnoise_allmasks.fits"
    integer(kind=Int32), dimension(:,:), pointer :: total_arrayP
    character(len=1024), dimension(1024) :: file
    character(256)      :: expression, finalExpression
    integer(kind=Int32) :: ccdNo
    real(kind=double)   :: nevt_cut
    logical(kind=bool)  :: savemasks, identifynoisyframes, applyfilter
    integer(kind=Int32), dimension(2)   :: imageDim

    character(len=13), parameter :: CodeVersion = '0.7.4'


    write(errstr,*) 'EPNOISE ',trim(CodeVersion),' is running...'
    call message(AppMsg,Sparse)
   
    applyfilter = booleanParameter('applyfilter')
    savemasks = booleanParameter('savemasks')
    identifynoisyframes = booleanParameter('identifynoisyframes')

    nfiles = parameterCount('set')
    write(errstr,'(a,t50,i10.0)')'Number of files to be processed : ',nfiles
    call message(AppMsg,Sparse)
    if (nfiles == 0) then
       call error("MissingParameter", "Missing input file name")
    endif

    finalExpression = ''
    if ((savemasks).and.(identifynoisyframes)) then
      imageDim(1) = mxdim
      imageDim(2) = mydim
      outmasks = dataSet(outmapset,CREATE,USE_ENVIRONMENT)
!     call addAttributes(attributable(outmasks),attributable(outfile))
      write(outMapExtension,'(a)') 'ALLMASKS'
      total = addArray(outmasks, outMapExtension, Integer32, imageDim)
      total_arrayP => int32Array2Data(total)
    endif

    do i = 1,nfiles
        file(i) = trim(stringParameter('set',i-1))
        call message(AppMsg,Sparse,' ')
        write(errstr,'(a,a,a)') 'Working on file: ',trim(file(i)),' ...'
        call message(AppMsg,Sparse)

        if (identifynoisyframes) then
            call EPN_identifynoisyframes(file(i),savemasks)
        endif

        if (applyfilter) then

            call removenoisyframes(trim(file(i)),nevt_cut,ccdNo)

            write(expression,'(a,f4.1)') ' !( (NEVT_FRM >=' ,nevt_cut
            expression = trim(expression)//') && (CCDNR == '
            write(expression,'(a,i2.0)') trim(expression) ,ccdNo
            if (i == nfiles) then
                expression = trim(expression)//' ) ) '
            else
                expression = trim(expression)//' ) ) && '
            endif
            finalExpression = trim(finalExpression)//trim(expression)
        endif
    enddo

    if ((savemasks).and.(identifynoisyframes)) then
       call message(AppMsg,Sparse,'total_arrayP = total_array ...')
       total_arrayP = total_array
       call message(AppMsg,Sparse,'call release(total)...')
       call release(total)
       call message(AppMsg,Sparse,'call release(outmasks)...')
       call release(outmasks)
       call message(AppMsg,Sparse,'Done.')
    endif

    if (applyfilter) then
        write(errstr,'(a,t50,a)')'Clean events expression : ', expression
        call message(AppMsg,Sparse)
        call filterEventFile(expression)
    endif

    if (allocated(ind)) deallocate(ind)

  end subroutine epnoise


!===================================================================
! Create the NEVT_FRM column and write keywords.
! It also calls to the mask builder subroutine
!
!====================================================================
    subroutine EPN_identifynoisyframes(file_name,savemasks)
        implicit none

        character(len=*), intent(IN) :: file_name
        logical(kind=bool), intent(IN) :: savemasks

        type(DataSetT) :: inSet
        type(TableT)   :: tab_events
        type(ColumnT)  :: nevt_frm_col, rawX, rawY, pha, frame

        real(kind=double)   :: dl0, dnorm0
        integer(kind=Int32) :: k0,k1
        integer(kind=Int16), dimension(:), pointer :: nevt_frm

        integer(kind=Int32), dimension(1)  :: dims
        integer(kind=Int32) :: i, k,j, jfr, ccdNo
        integer            ::  ix, iy, jx, jy, jx0, jy0
        integer             :: noisecut
        integer(kind=Int32) :: ix0(1:12), iy0(1:12), sign
        data ix0 /195,130, 65,261,326,391,196,261,326,130, 65,  0/
        data iy0 /403,403,403,403,403,403,  0,  0,  0,  0,  0,  0/

        call message(AppMsg,Sparse,'Calculating noisy frames ...')

        noisecut = intparameter("noisecut")

        !--------
        ! Open raw event file.
        !-------
        inSet = dataSet(file_name, MODIFY, USE_ENVIRONMENT)
        tab_events = table(inSet,"EVENTS")
        nrows = numberOfRows(tab_events)
        ccdNo = getCcd(block(tab_events))

        !--------
        ! Read columns of interest
        !--------

        frame =  column(tab_events,'FRAME',READ)
        frameCol => int32Data(frame)

        rawX  =  column(tab_events,'RAWX',READ)
        rawXCol => int16Data(rawX)
        rawY = column(tab_events,'RAWY',READ)
        rawYCol => int16Data(rawY)

        if (hasColumn(tab_events,'ENERGY')) then
            pha = column(tab_events,"ENERGY",READ)
        else
            pha = column(tab_events,"PHA",READ)
        endif
        energyCol => int16Data(pha)

        !--------
        ! Determine the frame start values
        !--------

        jfr = -1
        k   =  0
        if (nrows > 0) then
            if (allocated(ind)) deallocate(ind)
            allocate(ind(1:nrows))
            do i = 1, nrows
                if (frameCol(i-1) /= jfr) then
                    if (k < maxfrm) then
                        k = k + 1
                        ind(k) = i
                    else
                        call error('TooManyFrames','Too many frames')
                    endif
                    jfr = frameCol(i-1)
                endif
            enddo
        else
            call error('RawEventFileEmpty', 'epframes output file is empty')
        endif

        nfrm = k

        write(errstr,'(a,t50,i10)') 'Total number of frames : ',nfrm
        call message(AppMsg,Sparse)

        !-------------------------------------------------------
        ! get the mask (convention: 1 = opaque, 0 = transparent)
        !-------------------------------------------------------
        do j = 1, nydim
            do i = 1, nxdim
                i4mask(i,j) = 0
            enddo
        enddo


        !-----------------------------------------------------------------
        ! initialize the NEVT_FRM histogram:
        !-----------------------------------------------------------------

        do i = 1, maxhst
            i4hst(i) = 0.
        enddo
        !
        !     initialize the output map:
        !
        do j = 1, nydim
            do i = 1, nxdim
                i4map(i,j) = 0
            enddo
        enddo

        !-----------------------------------------------------------
        ! Initialize the NEVT_FRM column
        !------------------------------------------------------------
        dims(1)=1
        if (hasColumn(tab_events,'NEVT_FRM')) then
            call message(AppMsg,Sparse,'NEVT_FRM column already exists. Overwriting ...')
            nevt_frm_col =  column(tab_events,'NEVT_FRM',MODIFY)
        else
            call message(AppMsg,Sparse,'NEVT_FRM does not exist. Creating column ...')
            nevt_frm_col    = addColumn(tab_events,"NEVT_FRM",Integer16,"",dims, &
            comment="Number of events per Frame")
        endif

        nevt_frm => int16Data(nevt_frm_col)
        nevt_frm(0:nrows-1) = 0

        !---------------------
        ! Subroutine that calculates the number of events per frame and CCD
        !---------------------
        ! maximum amplitude of events to be considered.
        ! To get all potential sources
        iamx = 30000
        call framecounter(nevt_frm)


        ! -----------------------------------------------------------------
        !  fit poissonian distribution
        ! -----------------------------------------------------------------
        call ftpois(60,i4hst,dl0,dnorm0,k0,k1)

        write(errstr,'(a,t50,f10.3)') 'LAMBDA (lambda of poissonian fit) :',dl0
        call message(AppMsg,Sparse)
        write(errstr,'(a,t50,f10.3)') 'NORM (normalization of poissonian fit) :',dnorm0
        call message(AppMsg,Sparse)
        write(errstr,'(a,t50,i10.0)') 'NEVT_CUT (suggested cut value ("10%")) :',k0
        call message(AppMsg,Sparse)
        write(errstr,'(a,t50,i10.0)') 'NEVT_ALT (alternative cut value (1.0)) :',k1
        call message(AppMsg,Sparse)


        !----------------------------------------------------------------------
        ! Create mask (Bright sources and bad pixels)
        !----------------------------------------------------------------------
        call message(AppMsg,Sparse,'Creating mask (Bad pixel and bright sources) ...')
        call epnoisemask(inSet,i4map,i4mask,savemasks,ccdNo)


        if (savemasks) then
          if (ccdNo .ge.7) then
            sign = +1
          else
            sign = -1
          endif
          jx0 = ix0(CCDNo)
          jy0 = iy0(CCDNo)
          write(errstr,'(a,3i4,sp,i3)') 'MASK offsets: ',CCDNo,jx0,jy0,sign
          call message(AppMsg,Sparse)
          do iy = 1,nydim
            jy = jy0+sign*iy
            do ix = 1,nxdim
              jx = jx0+sign*ix
              total_array(jx,jy) = i4mask(ix,iy)
            enddo
          enddo
        endif

        !----------------------------------------------------------------------------------
        ! flag noise events/frames via external parameter instead of internal calculation
        ! which apparently returns more or less random numbers due to probably strange
        ! memory corruption caused by improper code
        !----------------------------------------------------------------------------------

        k0 = noisecut   ! epnoise-0.7

!        k = 0
!        do j = 1, nydim
!            do i = 1, nxdim
!                if (i4mask(i,j).eq.0)  k = k + 1
!            enddo
!        enddo


        !-----------------------------------------------------------------
        !
        !-----------------------------------------------------------------
        !     initialize the NEVT_FRM histogram:
        !
        do i = 1, maxhst
            i4hst(i) = 0.
        enddo
        !
        !     initialize the output map:
        !
        do j = 1, nydim
            do i = 1, nxdim
                i4map(i,j) = 0
            enddo
        enddo

        !-----------------------------------------------------------
        !
        !------------------------------------------------------------
        iamx = 30
        call framecounter(nevt_frm)

        ! -----------------------------------------------------------------
        !  fit poissonian distribution
        ! -----------------------------------------------------------------

        call ftpois(60,i4hst,dl0,dnorm0,k0,k1)

        call setAttribute(tab_events,"LAMBDA",dl0,"","lambda of poissonian fit")
        call setAttribute(tab_events,"NORM",dnorm0,"","normalization of poissonian fit")
        call setAttribute(tab_events,"NEVT_CUT",k0,"","suggested cut value (10%)")
        call setAttribute(tab_events,"NEVT_ALT",k1,"","alternative cut value (1.0)")

        write(errstr,'(a,t50,f10.3)') 'LAMBDA (lambda of poissonian fit) :',dl0
        call message(AppMsg,Sparse)
        write(errstr,'(a,t50,f10.3)') 'NORM (normalization of poissonian fit) :',dnorm0
        call message(AppMsg,Sparse)
        write(errstr,'(a,t50,i10.0)') 'NEVT_CUT (suggested cut value ("10%")) :',k0
        call message(AppMsg,Sparse)
        write(errstr,'(a,t50,i10.0)') 'NEVT_ALT (alternative cut value (1.0)) :',k1
        call message(AppMsg,Sparse)

        !-------------------------
        !
        !-------------------------
        call release(nevt_frm_col)
        call release(pha)
        call release(rawY)
        call release(frame)
        call release(rawX)
        call release(tab_events)
        call release(inSet)

        !-----------------------------------------------------------------
        ! perform some statistical calculations:

        i4ac1(1) = i4hst(maxhst)
        do i = maxhst-1, 1, -1
            i4ac1(i) = i4ac1(i+1) + i4hst(i)
        enddo
        !
        i4ac2(1) = i4hst(maxhst)
        do i = maxhst-1, 1, -1
            i4ac2(i) = i4ac2(i+1) + i4hst(i)*(i-1)
        enddo
        !
        i4ac3(1) = i4hst(1)
        do i = 2, maxhst
            i4ac3(i) = i4ac3(i-1) + i4hst(i)
        enddo
        !
        i4ac4(1) = 0
        do i = 2, maxhst
            i4ac4(i) = i4ac4(i-1) + i4hst(i)*(i-1)
        enddo
        !
        call writeInfo(k0)

    end subroutine EPN_identifynoisyframes
!====================================================================




!=====================================================================
! Calculate the number of events per frame
!
!=====================================================================
    subroutine framecounter(nevt_frm)
        implicit none
        integer(kind=Int16),  dimension(:), pointer :: nevt_frm !intent(inout),
        integer(kind=Int32) :: i,row,ifrm,nbuf,irow,k
        integer(kind=Int16) :: kk


        row = 0
        do ifrm = 1, nfrm

            irow = ind(ifrm)
            if (ldebug) then
              write(errstr,*) 'ifrm,irow=',ifrm,irow
              call message(AppMsg,Noisy)
            endif

            if (ifrm.lt.nfrm) then
                nbuf = ind(ifrm+1) - irow
            else
                nbuf = nrows - irow + 1
            endif

            k = 0

            do i = 1,nbuf

                !write(*,*)'energyCol( ', (i-1)+row , ' ) ',energyCol((i-1)+row)
                if (energyCol((i-1)+row).le.iamx) then

                    kx = rawXCol((i-1)+row)
                    ky = rawYCol((i-1)+row)
                    if (ldebug) then
                      write(errstr,*) "i,row,kx,ky=",i,row,kx,ky
                      call message(AppMsg,Noisy)
                    endif
                    if (i4mask(kx,ky).eq.0) then
                        k = k + 1
                        i4map(kx,ky) = i4map(kx,ky) + 1
                    endif
                endif
            enddo

            if (k .gt. 32767) then
              write(errstr,*) 'Framecounter() variable K too large: K = ',k
              call message(AppMsg,0)
              kk = k
            else
              kk = k
            endif
            do i = 1,nbuf
                noise(i) = kk
                nevt_frm((i-1)+row) = kk
            enddo

            row = row + nbuf

            k = k + 1
            if(k.le.maxhst) i4hst(k) = i4hst(k) + 1

        enddo

    end subroutine framecounter
!=====================================================================


!=====================================================================
! Write statistical information
!
!=====================================================================
    subroutine writeInfo(k0)

    implicit none

    integer(kind=Int32),intent(in) :: k0
    integer(kind=Int32)            :: k
    real(kind=double)              :: h1,h2

!   write(*,*) 'K0 = ',k0
    write(errstr,'(a,i10)') "Consequences of removing frames with NEVT_FRM >  ",k0-1
    call message(AppMsg,Sparse)
    k = nfrm - i4ac1(k0+1)

    write(errstr,'(a,i8,a,i7,2x,a,f6.2,a)') '    Number of frames: ',nfrm,' --> ',k,&
        '(',real(k)/real(nfrm)*100.0,'%)'
    call message(AppMsg,Sparse)
    k = i4ac2(1) - i4ac2(int(k0)+1)

    write(errstr,'(a,i10.0,a,i7,2x,a,f6.2,a)') '    "noise" events: ',i4ac2(1),' --> ',k,&
           '(',real(k)/real(i4ac2(1))*100.0,'%)'
    call message(AppMsg,Sparse)
    h1 = dble(i4ac4(maxhst))/dble(i4ac3(maxhst))
    h2 = dble(i4ac4(k0))/dble(i4ac3(k0))

    write(errstr,'(a,f7.3,a)') '    before:            ',h1,' "noise" events / frame'
    call message(AppMsg,Sparse)
    write(errstr,'(a,f7.3,a)') '    afterwards:        ',h2,' "noise" events / frame'
    call message(AppMsg,Sparse)
    k = nfrm - i4ac1(k0+1)
    write(errstr,'(a,f6.2,a,2x,a,f6.2,a)')'--> "noise" rate reduction: ',(1.0-h2/h1)*100.0,'%',&
      ' (loss of exposure: ',dble(nfrm-k)/dble(nfrm)*100.0,'%)'
    call message(AppMsg,Sparse)


    end subroutine writeInfo


!=====================================================================





!======================================================================
    subroutine ftpois(n,i4hst,dl0,dnorm0,k0,k1)
!
!     fits a poissonian distribution to i4hst(1..n)
!
    implicit none
!
    integer, intent(in)              :: n
    integer(kind=Int32), intent(in)  :: i4hst(n)
    integer(kind=Int32), intent(out) :: k0, k1
    real(kind=double), intent(out)   :: dl0, dnorm0
    real(kind=double)   :: dl, dnorm, dh, dchi2, dflux, dchi20
    real(kind=double)   :: flux
    integer             :: i,k, iloop
!
      k = 0
      do i = 1, n
        k = k + i4hst(i)
      enddo
      flux = dble(k)
!
      dl0    = 0.d0
      dnorm0 = 0.d0
      dchi20 = 1.e10
!
      do iloop = 100, 4000, 1
        dl = iloop*0.001d0
        dh = 0.d0
        do i = 1, n-1
          dh = dh + dpoiss(dl,i)
        enddo
        dflux = dh
!
        dnorm = flux / dflux
!
        dchi2 = 0.d0
        do i = 2, n
          dh = dble(i4hst(i)) - dpoiss(dl,i-1) * dnorm
          dchi2 = dchi2 + dh**2 / dble(max(i4hst(i),1))
        enddo
!
        dchi2 = dchi2 / dble(n)
!
        if (dchi2.lt.dchi20) then
          dl0 = dl
          dnorm0 = dnorm
          dchi20 = dchi2
        endif
      enddo
!
!     determine where fit value drops below 10% of histogram:
!
      k = 1
      do while (dpoiss(dl0,k)*dnorm0.gt.0.1d0*dble(i4hst(k+1)) &
               .and. k.lt.(n-1))
        k = k + 1
      enddo
      k0 = k
!
!     determine where fit value drops below 1.0:
!
      k = 1
      do while (dpoiss(dl0,k)*dnorm0.gt.1.d0 .and. k.lt.(n-1))
        k = k + 1
      enddo
      k1 = k
!
      end subroutine ftpois


!======================================================================

    real(double) function dpoiss(dl,k)
!
    implicit none
    real(kind=double),intent(in) :: dl
    integer, intent(in)          :: k

!
    dpoiss = dl**k / factrl(k) * dexp(-dl)
!
    end function dpoiss


!======================================================================
!
!
!
!======================================================================
     real(double) function factrl(n)
!
! adapted from NR
!
     integer, intent(in) ::  n

     integer(kind=Int32) ::  j,ntop
     real(kind=double)   ::  a(33),aa
     save ntop,a
     data ntop,a(1)/0,1.d0/
!
     if (n.lt.0) then
!-mjf      pause 'negative factorial in factrl'
!-mjf ! cannot happen as all calls to factrl and dpoiss are with positive argument
       print*,'negative factorial in factrl'
     else if (n.le.ntop) then
       factrl = a(n+1)
     else if (n.le.32) then
       do j = ntop+1,n
         a(j+1) = j * a(j)
       enddo
       ntop = n
       factrl = a(n+1)
     else
       aa = dble(n+1)
       factrl = exp(gammln(aa))
     endif
!
     end function factrl

!======================================================================
!
!
!
!======================================================================
      real(double) function gammln(xx)
!
! adapted from NR 2nd (!!) edition
!
      real(kind=double), intent(in) :: xx
      integer(kind=Int32)           :: j
      real(kind=double)             :: ser,stp,tmp,x,y,cof(6)
!
      save cof,stp
!
      data cof,stp/76.18009172947146d0,-86.50532032941677d0, &
       24.01409824083091d0,-1.231739572450155d0,0.1208650973866179d-2, &
       -0.5395239384953d-5,2.5066282746310005d0/
!
      x = xx
      y = x
      tmp = x + 5.5d0
      tmp = (x + 0.5d0) * log(tmp) - tmp
      ser = 1.000000000190015d0
      do j = 1,6
        y = y + 1.d0
        ser = ser + cof(j) / y
      enddo
      gammln=tmp+log(stp*ser/x)
!
      end function gammln
!======================================================================



end module epnoise_mod
