!+
! Module: updateexpo_mod
!
! Authors:  Aitor Ibarra - XMM-Newton SOC
!
! Description: Algorithm to reject soft X-ray noise in the EPIC-pn camera
! Module in charge of create new GTI table and update the ONTIME and LIVETIME keywords.
!
!
! Comments:
!
!
! $Id: updateexpo_mod.f90,v 1.2 2011/09/13 14:07:42 aibarra Exp $
!-

module updateexpo_mod
    use dal
    use param
    use errorHandling
    use dal_aux
    use datafilter
    use array_utils
    use caltypes
    use dss_aux

    use intervals_aux

    implicit none

  type :: frameType
    real(single)    :: fracExp, deadTimeFraction
    real(double)    :: time
    type(IntervalT) :: interval
    logical(bool)   :: mask
  end type frameType
  integer, parameter :: targetCCD = 4 ! Central CCD for EPIC-pn

contains

    subroutine updateexpo(tevents,ccdNo)

    implicit none
    type(DataSetT) :: set
    type(TableT) :: exposuTab, tabgti
    type(TableT), intent(in) :: tevents
    character(256) :: exposuTabName
    integer(int32), intent(in) :: ccdNo
    integer(int32) :: numFrames,numEvents,i,j,il, frm, nfrm,frmcounter, ntgti, &
            sizeGti,newGtis, totalFrames
    type(frameType), allocatable :: frames(:)
    type(IntervalT), pointer :: gtis(:) => null()
    real(double), pointer :: evtime(:) => null()
    real(double) :: livetime
    integer(kind=Int32), dimension(:), allocatable :: evtfrm
    real(kind=double) :: exposuTimeDel,timeInBinGtis, ontime, FC0
    real(kind=double),    pointer :: startColGti(:), stopColGti(:)
    type(IntervalT), pointer :: gti(:), andedGti(:) => null()
    type(BlockT)   :: eventBlock
    type(ColumnT) :: startCol, stopCol
    real(kind=double), dimension(:), pointer :: startColData, stopColData
    character(len=8) :: ccdLiveTime,ccdOntime,stdGTIext


    write(exposuTabName,'(a6,i2.2)') "EXPOSU" ,ccdNo

    write(errstr,'(a,t50,a)')'Reading exposure extension : ',exposuTabName
    call message(AppMsg,Sparse)

    set = dataSet(trim(stringParameter('eventset')), MODIFY)
    exposuTab = table(set, exposuTabName)

    numFrames = numberOfRows(exposuTab)
    allocate(frames(numFrames))
    call readColDataToFixed(exposuTab, 'TIME',    frames%time)

    exposuTimeDel = real64Attribute(exposuTab, 'TIMEDEL')
    FC0 = real64Attribute(exposuTab, 'AX_FC0')
    call readColDataToFixed(exposuTab, 'FRACEXP', frames%fracExp)

    frames%interval%lower = frames%time - exposuTimeDel / 2.0d0
    frames%interval%upper  = frames%time + exposuTimeDel / 2.0d0
    write(errstr,'(a,t50,i10)')'Total number of frames : ',size(frames%interval)
    call message(AppMsg,Sparse)

    numEvents = numberOfRows(tevents)
    write(errstr,'(a,t50,i10.0)')'noisy events : ',numEvents
    call message(AppMsg,Sparse)
    allocate(evtime(0:numEvents-1))
    j = 1
    i = 0
    frm = 0
    nfrm = 0
    frmcounter = 1
    totalFrames = 1
    allocate(evtfrm(1:numEvents))
    allocate(gtis(numEvents))
    gtis%lower = 0.0
    gtis%upper = 0.0
    evtfrm = 0
    call readColDataToFixed(tevents, 'TIME', evtime)

    ! Loop that identifies the frame for all the noisy events
    ! Then, it creates the corresponding GTIs
    gtis(frmcounter)%lower = frames(j)%interval%lower
    do while ( i < numEvents )
       if ( (evtime(i) >= frames(j)%interval%lower) .and. (evtime(i) < frames(j)%interval%upper) ) then
            !write(*,*) 'EVENT( ', i ,') ',evtime(i), ' Low( ', j , ') ', frames(j)%interval%lower, &
            !    ' High( ', j , ') ', frames(j)%interval%upper
            i = i + 1
            frm = j
            if (nfrm /= frm) then

                !write(*,*)'NEW FRM ', frm
                evtfrm(frmcounter) = frm
                !
                gtis(frmcounter)%upper = frames(j)%interval%lower
                frmcounter = frmcounter+1
                gtis(frmcounter)%lower = frames(j)%interval%upper

                nfrm = frm
            endif
           ! j = 0
       endif
       if ( (i < numEvents) ) then
            if (evtime(i) >= frames(j)%interval%upper) then
                j = j + 1
            endif
       endif
    enddo
    gtis(frmcounter)%upper = frames(numFrames)%interval%upper
    write(errstr,'(a,t50,i10.0)')'Frames to be removed : ',frmcounter
    call message(AppMsg,Sparse)
!===============
! GTI update
!===============
    newGtis = size(gtis)


    write(stdGTIext,'(a6,i2.2)') "STDGTI" ,ccdNo
    write(errstr,'(a,a)')'Reading STD Extension ',stdGTIext
    call message(AppMsg,Sparse)
    tabgti = table(set,stdGTIext)
    ntgti = numberOfRows(tabgti)
    eventBlock = block(set, 'EVENTS')

    call readColDataToPtr(tabgti, 'START', startColGti)  ! defined in dal_aux

    call readColDataToPtr(tabgti, 'STOP',  stopColGti)   ! defined in dal_aux

    call readGtisFromDss(set, eventBlock, ccdNo, gti)    ! CCD Chip No.X

    sizeGti = size(gti)


    newGtis = size(gtis)

!====
! Insert ROWS!!!!
!===
    write(errstr,'(a,t50,i10.0)')'GTIs from noisy frames : ', size(gtis)
    call message(AppMsg,Sparse)

    write(errstr,'(a,t50,i10.0)')'STDGTI size : ',size(gti)
    call message(AppMsg,Sparse)


    call andIntervals(gtis,gti, andedGti)

    write(errstr,'(a,t50,i10.0)')'Final GTI size : ', size(andedGti)
    call message(AppMsg,Sparse)


    timeInBinGtis = sum(andedGti%upper) - sum(andedGti%lower)

    write(errstr,'(a,t50,f12.4)')'ONTIME (Noisy-filtered) : ',(sum(gti%upper) - sum(gti%lower))-timeInBinGtis
    call message(AppMsg,Sparse)

    write(errstr,'(a,t50,i12.1)')'Inserting new GTI Rows : ', size(andedGti)
    call message(AppMsg,Sparse)
    call insertRows( tabgti,  numberOfRows(tabgti), size(andedGti)-numberOfRows(tabgti) )
    write(errstr,'(a,t50,i10.1)')'Rows added to the final STDGTI extension : ',numberOfRows(tabgti)
    call message(AppMsg,Sparse)
!====
! Insert ROWS!!!!
!===


    startCol =  column(tabgti,'START')
    startColData => real64Data(startCol)
    do i = 0, size(andedGti)-1
        startColData(i) = andedGti( i+1 )%lower
    enddo

    stopCol =  column(tabgti,'STOP',MODIFY)
    stopColData => real64Data(stopCol)
    do i = 0, size(andedGti)-1
        stopColData(i) = andedGti( i+1 )%upper
    enddo
    !call setData( startCol, sizeGti-1, gtis( 1 : frmcounter )%lower)
    !call setData( stopCol, sizeGti-1, gtis( 1 : frmcounter )%upper)

    !! ONTIME
    ontime = 0
    do il = 0, size(andedGti)-1
      ontime = ontime + stopColData(il)-startColData(il)
    enddo
    call setAttribute(exposuTab, "ONTIME", ontime, "s", &
               "Total good times")
    ! Set the value to the targetCCD
    if (ccdNo == targetCCD) then
        call setAttribute(tevents, "ONTIME", livetime, "s", &
               "Sum of GTIs for the central CCD")
    endif


    write(ccdOntime,'(a6,i2.2)') "ONTIME" ,ccdNo
    call setAttribute(tevents, ccdOntime, ontime, "s", &
               "Sum of GTIs for a particular CCD")

    !!
    call release (startCol)
    call release (stopCol)
    deallocate(andedGti)
    call release (tabgti)
    nullify(gtis)
    deallocate(gti)
!===============
! GTI update
!===============

!===============
! LIVETIME and EXPOSURE update
!===============

    livetime = 0.0d0
    livetime = max(ontime*FC0,0.0d0)
    write(errstr,'(a,t50,f10.2)')'LIVETIME (unfiltered-filtered) : ', real64Attribute(exposuTab, 'LIVETIME') - livetime
    call message(AppMsg,Sparse)
    call setAttribute(exposuTab, "LIVETIME", livetime, "s", &
               "Good times * FC0")

    call setAttribute(exposuTab, "EXPOSURE", livetime, "s", &
               "temporary = LIVETIME")

    ! Set the value to the targetCCD
    if (ccdNo == targetCCD) then
        call setAttribute(tevents, "LIVETIME", livetime, "s", &
               "Live time for the central CCD")
    endif

    write(ccdLiveTime,'(a6,i2.2)') "LIVETI" ,ccdNo
    call setAttribute(tevents, ccdLiveTime, livetime, "s", &
               "Live time for a particular CCD")
    deallocate(evtime,frames,evtfrm)

    call release(set)

    end subroutine updateexpo


!==============================
!
!=============================
    subroutine updateLiveTime(exposuTab,livetime)
    implicit none

    real(double), intent(inout) :: livetime
    type(TableT), intent(in) :: exposuTab
    type(ColumnT) :: cfrac
    real(kind=single), dimension(:), pointer :: xxfrac
    real(kind=double) :: exposuTimeDel
    integer(int32) :: numFrames,i

    exposuTimeDel = real64Attribute(exposuTab, 'TIMEDEL')
    cfrac = column(exposuTab,"FRACEXP", Read)
    xxfrac => real32Data(cfrac)
    numFrames = numberOfRows(exposuTab)


    do i = 0, numFrames-1
       livetime = livetime + xxfrac(i)*exposuTimeDel
    enddo


    call release(cfrac)

    end subroutine updateLiveTime

end module updateexpo_mod
